import random

def euclideanDist(p1, p2):
    dist = 0.0
    for i in range(len(p1.coord)):
        dist += ((p1.coord[i] - p2.coord[i]) * (p1.coord[i] - p2.coord[i]))
    return dist

def manhattanDist(p1, p2):
    dist = 0.0
    for i in range(len(p1.coord)):
        dist += abs(p1.coord[i] - p2.coord[i])
    return dist

def addNeighbours(p1, p2):
    if (not (p2.id in p1.neighDic)):
        p1.neighList.append(p2)
        p1.neighDic[p2.id] = 1
        
    if (not (p1.id in p2.neighDic)):
        p2.neighList.append(p1)
        p2.neighDic[p1.id] = 1

def addTreeNeighbours(p1, p2):
    if (not (p2.id in p1.treeNeighDic)):
        p1.treeNeighList.append(p2)
        p1.treeNeighDic[p2.id] = 1
        
    if (not (p1.id in p2.treeNeighDic)):
        p2.treeNeighList.append(p1)
        p2.treeNeighDic[p1.id] = 1

def dfs(p):
    global visited, numVisited

    visited[p.id] = 1
    numVisited += 1

    for q in p.treeNeighList:
        if (not (q.id in visited)):
            dfs(q)

def dfs2(p):
    global visited

    visited[p.id] = 1
    dmax = -1

    for q in p.treeNeighList:
        if (not (q.id in visited)):
            d = dfs2(q)
            if (d > dmax):
                dmax = d
            
    return dmax + 1

def constructP2PTopology():
    global all
    # construct the peer-to-peer geometric overlay topology        
    for p in all:
        allWithoutP = list(all)
        allWithoutP.remove(p)
        p.selectNeighbours(allWithoutP)

    return
    
    for idx in range(1, NUM_PEERS):
        p = all[idx]
        pidx = random.randint(0, idx - 1)
        addNeighbours(p, all[pidx])

        niter = 0
        exchange = 1
        while (exchange > 0):
            exchange = 0
            niter += 1
    
            for q in all[:idx+1]:
                q.extendedNeighboursList = q.getExtendedNeighbours()
                #q.extendedNeighboursList = all[:idx + 1]
                q.extendedNeighboursList.remove(q)

            for q in all[:idx+1]:
                q.prevNeighDic = q.neighDic
                q.cntPrevNeigh = len(q.neighList)
                q.neighDic = {}
                q.neighList = []

            for q in all[:idx+1]:
                q.selectNeighbours(q.extendedNeighboursList)
                 
            for q in all[:idx+1]:
                if (len(q.neighList) != q.cntPrevNeigh):
                    exchange = 1
                    break
                else:
                    for p in q.neighList:
                        if (not (p.id in q.prevNeighDic.keys())):
                            exchange = 1
                            break

                    if (exchange == 1):
                        break

            #print all[:idx+1]

        if (idx % 100 == 0):
            print "Finished with", idx, "in", niter, "iterations"

class Peer:
    def __init__(self, id, coord):
        self.id = id
        self.coord = coord
        self.neighList = []
        self.neighDic = {}
        self.treeNeighList = []
        self.treeNeighDic = {}

    def quadrant(self, peerList, dfunc):
        qDic = {}
        for val in range(1 << D):
            qDic[val] = []
            
        for p in peerList:
            v = 0
            for i in range(len(self.coord)):
                v = v << 1
                if (p.coord[i] < self.coord[i]):
                    v += 1
            qDic[v].append(p)
            
        for val in range(1 << D):
            ldist = []
            for p in qDic[val]:
                ldist.append((dfunc(self, p), p))
            ldist.sort()
            
            j = 0
            while ((j < K) and (j < len(ldist))):
                (d, p) = ldist[j]
                addNeighbours(self, p)
                j += 1

    def emptyBoundingBoxes(self, peerList):
        for p in peerList:
            xmin = []
            xmax = []
            for i in range(len(self.coord)):
                if (p.coord[i] < self.coord[i]):
                    xmin.append(p.coord[i])
                else:
                    xmin.append(self.coord[i])

                if (p.coord[i] > self.coord[i]):
                    xmax.append(p.coord[i])
                else:
                    xmax.append(self.coord[i])

            pok = 1
            for q in peerList:
                if (q.id != p.id):
                    qok = 1
                    for i in range(len(self.coord)):
                        if ((q.coord[i] < xmin[i]) or (q.coord[i] > xmax[i])):
                            qok = 0
                            break
                    
                    if (qok == 1):
                        pok = 0
                        break
            
            if (pok == 1):
                addNeighbours(self, p)
                        
    def selectNeighbours(self, peerList):
        if (NEIGH_SEL_METHOD == 1):
            self.quadrant(peerList, euclideanDist)
        elif (NEIGH_SEL_METHOD == 2):
            self.emptyBoundingBoxes(peerList)
            #self.quadrant(peerList, manhattanDist)

    def getExtendedNeighbours(self):
        enList = list(self.neighList)
        enDic = {}
        for p in enList:
            enDic[p.id] = 1
            
        enList2 = list(enList)
        for p in enList2:
            for q in p.neighList:
                if (not (q.id in enDic.keys())):
                    enList.append(q)
                    enDic[q.id] = 1
        
        # the neighbours of the neighbours
        return enList

    def multicast(self, zone):
        global cnt, visited
        
        #visited[self.id] = 1
               
        candNeighDic = {}
        for val in range(1 << D):
            candNeighDic[val] = []

        for p in self.neighList:
            ok = 1
            for j in range(D):
                vmin, vmax = zone[j]
                if ((p.coord[j] <= vmin) or (p.coord[j] >= vmax)):
                    ok = 0
                    
            if (ok == 1):
                v = 0
                for i in range(len(self.coord)):
                    v = v << 1
                    if (p.coord[i] < self.coord[i]):
                        v += 1
                candNeighDic[v].append(p)

        if (NEIGH_SEL_METHOD == 1):
            dfunc = euclideanDist
        elif (NEIGH_SEL_METHOD == 2):
            dfunc = manhattanDist
        
        zoneDic = {}
        selNeighList = []
        for val in range(1 << D):
            ldist = []
            for p in candNeighDic[val]:
                ldist.append((dfunc(self, p), p))
            ldist.sort()
            
            N = len(ldist)
            if (N > 0):
                M = N / 2
                d, p = ldist[M]
                
                pzone = []
                for j in range(D):
                    vmin, vmax = zone[j]
                    if (p.coord[j] < self.coord[j]):
                        pzone.append((vmin, self.coord[j]))
                    else:
                        pzone.append((self.coord[j], vmax))
                
                zoneDic[p.id] = pzone
                selNeighList.append(p)
                addTreeNeighbours(self, p)
        
        # multicast the message further
        for p in selNeighList:
            #print "Sending from", self.id, "(coord=", self.coord, ", zone=", zone, ") to", p.id, "(zone=", zoneDic[p.id], ")"
            p.multicast(zoneDic[p.id])

#NUM_PEERS = 1000
MAX_COORD = 10000
NEIGH_SEL_METHOD = 2

# initialize the peers
random.seed(1000)
        
flog = open("mcast_geom_zones_log-x.csv", "w")

for D in range(2,3):

    #for K in range(1,2):
    NUM_PEERS = 5000
    while (NUM_PEERS <= 100000):
        #print "D=", D, "; K=", K
        print "NUM_PEERS=", NUM_PEERS
            
        all = []
        for i in range(NUM_PEERS):
            lc = []
            for j in range(D):
                lc.append(random.random() * MAX_COORD)   
            all.append(Peer(i, lc))
        
        # reset the neighbour lists and dictionaries
        for p in all:
            p.neighList = []
            p.treeNeighList = []
            p.neighDic = {}
            p.treeNeighDic = {}
        
        # construct the peer-to-peer geometric overlay topology        
        constructP2PTopology()
        
        maxP2PDeg = 0
        avgP2PDeg = 0
        for p in all:
            if (len(p.neighList) > maxP2PDeg):
                maxP2PDeg = len(p.neighList)
            avgP2PDeg += len(p.neighList)
            
        avgP2PDeg /= NUM_PEERS

        print "Maximum P2P Degree=", maxP2PDeg
        print "Average P2P Degree=", avgP2PDeg
    
        if (NUM_PEERS < 1000):
            NUM_PEERS += 100
        elif (NUM_PEERS < 10000):
            NUM_PEERS += 1000
        else:
            NUM_PEERS += 10000

        continue
        
        # construct a multicast tree starting from every node
        diamAvg = 0.0
        maxDiam = 0
        maxDegAvg = 0.0
        maxDeg = 0
        
        for root in all:
            for q in all:
                q.treeNeighList = []
                q.treeNeighDic = {}

            zone = []
            for d in range(D):
                zone.append((-1, MAX_COORD + 1))

            root.multicast(zone)
            visited = {}
            numVisited = 0
            
            dfs(root)
            #print "Number of visited nodes=", numVisited
                       
            if (numVisited == NUM_PEERS):
                visited = {}
                diam = dfs2(root)               
                if (diam > maxDiam):
                    maxDiam = diam
                diamAvg += diam
               
                maxDegree = 0
                for p in all:
                    if (len(p.treeNeighList) > maxDegree):
                        maxDegree = len(p.treeNeighList)
            
                if (maxDegree > maxDeg):
                    maxDeg = maxDegree
                maxDegAvg += maxDegree
            else:
                maxDeg = NUM_PEERS + 1
                maxDiam = NUM_PEERS + 1
                
            #break
                
        diamAvg /= NUM_PEERS
        maxDegAvg /= NUM_PEERS
    
        print "Longest Root-to-Leaf Path=", maxDiam
        print "Average Longest Root-to-Leaf Path=", diamAvg
        print "Maximum Degree=", maxDeg
        print "Average Maximum Degree=", maxDegAvg
            
        flog.write(str(D)+";"+str(K)+";"+str(maxDiam)+";"+str(diamAvg)+";"+str(maxDeg)+";"+str(maxDegAvg)+";"+str(maxP2PDeg)+";"+str(avgP2PDeg)+"\n")

flog.close()
