import random

NUM_PEERS = 1000
MAX_COORD = 10000

def euclideanDist(p1, p2):
    dist = 0.0
    for i in range(len(p1.coord)):
        dist += ((p1.coord[i] - p2.coord[i]) * (p1.coord[i] - p2.coord[i]))
    return dist

def addNeighbours(p1, p2):
    if (not (p2.id in p1.neighDic)):
        p1.neighList.append(p2)
        p1.neighDic[p2.id] = 1
        
    if (not (p1.id in p2.neighDic)):
        p2.neighList.append(p1)
        p2.neighDic[p1.id] = 1

def addTreeNeighbours(p1, p2):
    if (not (p2.id in p1.treeNeighDic)):
        p1.treeNeighList.append(p2)
        p1.treeNeighDic[p2.id] = 1
        
    if (not (p1.id in p2.treeNeighDic)):
        p2.treeNeighList.append(p1)
        p2.treeNeighDic[p1.id] = 1

def dfs(p):
    global visited, numVisited, numInversions

    visited[p.id] = 1
    numVisited += 1

    for q in p.treeNeighList:
        if (not (q.id in visited)):
            #print q.id, q.time, "son of", p.id, p.time

            if (q.time > p.time):
                numInversions += 1
                
            dfs(q)

def dfs2(p):
    global visited

    visited[p.id] = 1
    dmax = -1

    for q in p.treeNeighList:
        if (not (q.id in visited)):
            d = dfs2(q)
            if (d > dmax):
                dmax = d
            
    return dmax + 1

def constructP2PTopology():
    global all
    # construct the peer-to-peer geometric overlay topology        
    for idx in range(1, NUM_PEERS):
        p = all[idx]
        pidx = random.randint(0, idx - 1)
        addNeighbours(p, all[pidx])

        niter = 0
        exchange = 1
        while (exchange > 0):
            exchange = 0
            niter += 1
    
            for q in all[:idx+1]:
                q.extendedNeighboursList = q.getExtendedNeighbours()
                #q.extendedNeighboursList = all[:idx + 1]
                q.extendedNeighboursList.remove(q)

            for q in all[:idx+1]:
                q.prevNeighDic = q.neighDic
                q.cntPrevNeigh = len(q.neighList)
                q.neighDic = {}
                q.neighList = []

            for q in all[:idx+1]:
                q.selectNeighbours(q.extendedNeighboursList)
                 
            for q in all[:idx+1]:
                if (len(q.neighList) != q.cntPrevNeigh):
                    exchange = 1
                    break
                else:
                    for p in q.neighList:
                        if (not (p.id in q.prevNeighDic.keys())):
                            exchange = 1
                            break

                    if (exchange == 1):
                        break

            #print all[:idx+1]

        if (idx % 100 == 0):
            print "Finished with", idx, "in", niter, "iterations"
    #for p in all:
    #    allWithoutP = list(all)
    #    allWithoutP.remove(p)
    #    p.selectNeighbours(allWithoutP)

class Peer:
    def __init__(self, id, coord, time):
        self.id = id
        self.coord = coord
        self.time = time
        self.neighList = []
        self.neighDic = {}
        self.treeNeighList = []
        self.treeNeighDic = {}

    def quadrant(self, peerList, dfunc):
        qDic = {}
        for val in range(1 << D):
            qDic[val] = []
            
        for p in peerList:
            v = 0
            for i in range(len(self.coord)):
                v = v << 1
                if (p.coord[i] < self.coord[i]):
                    v += 1
            qDic[v].append(p)
            
        for val in range(1 << D):
            ldist = []
            for p in qDic[val]:
                ldist.append((dfunc(self, p), p))
            ldist.sort()
            
            j = 0
            while ((j < K) and (j < len(ldist))):
                (d, p) = ldist[j]
                addNeighbours(self, p)
                j += 1

    def selectNeighbours(self, peerList):               
        if (NEIGH_SEL_METHOD == 1):
            self.quadrant(peerList, euclideanDist)

    def getExtendedNeighbours(self):
        enList = list(self.neighList)
        enDic = {}
        for p in enList:
            enDic[p.id] = 1
            
        enList2 = list(enList)
        for p in enList2:
            for q in p.neighList:
                if (not (q.id in enDic.keys())):
                    enList.append(q)
                    enDic[q.id] = 1
        
        # the neighbours of the neighbours
        return enList

    def selectPreferredTreeNeighbour(self):
        cand = []
        for p in self.neighList:
            if ((p.time > self.time) or ((p.time == self.time) and (p.id > self.id))):
                cand.append(p)
               
        tmax = -1
        ptn = None
        
        for p in cand:
            if (p.time > tmax):
                tmax = p.time
                ptn = p
                
        if (ptn != None):
            addTreeNeighbours(self, ptn)
            #print self.id, self.time, "son of", ptn.id, ptn.time

NEIGH_SEL_METHOD = 1

# initialize the peers
random.seed(1000)
        
all = []
for i in range(NUM_PEERS):
    lc = []
    time = random.random() * MAX_COORD      
    lc.append(time)
    all.append(Peer(i, lc, time))

flog = open("mcast_stab_log-x.csv", "w")

for D in range(2,11):
    # add a new dimension to the coordinates of every peer
    for p in all:
        p.coord.append(random.random() * MAX_COORD)

    for K in range(1,51):
        print "D=", D, "; K=", K
            
        # reset the neighbour lists and dictionaries
        for p in all:
            p.neighList = []
            p.treeNeighList = []
            p.neighDic = {}
            p.treeNeighDic = {}
        
        constructP2PTopology()
        
        # construct the multicast tree
        for p in all:
            p.selectPreferredTreeNeighbour()
        
        # check the obtained tree        
        tmax = -1
        root = None
        
        for p in all:
            if (p.time > tmax):
                tmax = p.time
                root = p
        
        visited = {}
        numVisited = 0
        numInversions = 0
        
        dfs(root)
        
        print "Number of visited nodes=", numVisited
        print "Number of Inversions=", numInversions
        
        if (numVisited == NUM_PEERS):
            diam = 0
            for p in all:
                visited = {}
                lpath = dfs2(p)
                if (lpath > diam):
                    diam = lpath
                
            print "Tree Diameter=", diam
        
        maxDegree = 0
        for p in all:
            if (len(p.treeNeighList) > maxDegree):
                maxDegree = len(p.treeNeighList)
        
        print "Maximum Degree=", maxDegree
        
        flog.write(str(D)+";"+str(K)+";"+str(diam)+";"+str(maxDegree)+";"+str(numVisited)+";"+str(numInversions)+"\n")
        
        if (maxDegree == NUM_PEERS - 1):
            break

flog.close()
